from itertools import chain
from struct import Struct

from .base import Column
from .intcolumn import UInt64Column
from ..util.helpers import pairwise


class ArrayColumn(Column):
    """
    Nested arrays written in flatten form after information about their
    sizes (offsets really).
    One element of array of arrays can be represented as tree:
    (0 depth)          [[3, 4], [5, 6]]
                      |               |
    (1 depth)      [3, 4]           [5, 6]
                   |    |           |    |
    (leaf)        3     4          5     6

    Offsets (sizes) written in breadth-first search order. In example above
    following sequence of offset will be written: 4 -> 2 -> 4
    1) size of whole array: 4
    2) size of array 1 in depth=1: 2
    3) size of array 2 plus size of all array before in depth=1: 2 + 2 = 4

    After sizes info comes flatten data: 3 -> 4 -> 5 -> 6
    """
    py_types = (list, tuple)

    def __init__(self, nested_column, **kwargs):
        self.size_column = UInt64Column()
        self.nested_column = nested_column
        self._write_depth_0_size = True
        super(ArrayColumn, self).__init__(**kwargs)

    def write_data(self, data, buf):
        # Column of Array(T) is stored in "compact" format and passed to server
        # wrapped into another Array without size of wrapper array.
        self.nested_column = ArrayColumn(self.nested_column)
        self.nested_column.nullable = self.nullable
        self.nullable = False
        self._write_depth_0_size = False
        self._write(data, buf)

    def read_data(self, rows, buf):
        self.nested_column = ArrayColumn(self.nested_column)
        self.nested_column.nullable = self.nullable
        self.nullable = False
        return self._read(rows, buf)[0]

    def _write_sizes(self, value, buf):
        nulls_map = []

        column = self
        sizes = [len(value)] if self._write_depth_0_size else []

        while True:
            nested_column = column.nested_column
            if not isinstance(nested_column, ArrayColumn):
                if column.nullable:
                    nulls_map = [x is None for x in value]
                break

            offset = 0
            new_value = []
            for x in value:
                offset += len(x)
                sizes.append(offset)
                new_value.extend(x)

            value = new_value
            column = nested_column

        if nulls_map:
            self._write_nulls_map(nulls_map, buf)

        ns = Struct('<{}Q'.format(len(sizes)))
        buf.write(ns.pack(*sizes))

    def _write_data(self, value, buf):
        if self.nullable:
            value = value or []

        if isinstance(self.nested_column, ArrayColumn):
            value = list(chain.from_iterable(value))

        if value:
            self.nested_column._write_data(value, buf)

    def _write_nulls_data(self, value, buf):
        if self.nullable:
            value = value or []

        if isinstance(self.nested_column, ArrayColumn):
            value = list(chain.from_iterable(value))
            self.nested_column._write_nulls_data(value, buf)
        else:
            if self.nested_column.nullable:
                self.nested_column._write_nulls_map(value, buf)

    def _write(self, value, buf):
        self._write_sizes(value, buf)
        self._write_nulls_data(value, buf)
        self._write_data(value, buf)

    def read_state_prefix(self, buf):
        return self.nested_column.read_state_prefix(buf)

    def write_state_prefix(self, buf):
        self.nested_column.write_state_prefix(buf)

    def _read(self, size, buf):
        slices_series = [[0, size]]
        nested_column = self.nested_column

        cur_level_slice_size = size
        cur_level_slice = None
        while (isinstance(nested_column, ArrayColumn)):
            if cur_level_slice is None:
                cur_level_slice = [0]
            ns = Struct('<{}Q'.format(cur_level_slice_size))
            nested_sizes = ns.unpack(buf.read(ns.size))
            cur_level_slice.extend(nested_sizes)
            slices_series.append(cur_level_slice)
            cur_level_slice = None
            cur_level_slice_size = nested_sizes[-1] if len(nested_sizes) > 0 \
                else 0
            nested_column = nested_column.nested_column

        n_items = cur_level_slice_size if size > 0 else 0
        nulls_map = None
        if nested_column.nullable:
            nulls_map = self._read_nulls_map(n_items, buf)

        data = []
        if n_items:
            data = list(nested_column._read_data(
                n_items, buf, nulls_map=nulls_map
            ))

        # Build nested structure.
        for slices in reversed(slices_series):
            data = [data[begin:end] for begin, end in pairwise(slices)]

        return tuple(data)


def create_array_column(spec, column_by_spec_getter):
    inner = spec[6:-1]
    return ArrayColumn(column_by_spec_getter(inner))



import numpy as np

from ... import defines
from .base import NumpyColumn


class NumpyStringColumn(NumpyColumn):
    dtype = np.dtype('object')

    default_encoding = defines.STRINGS_ENCODING

    def __init__(self, encoding=default_encoding, **kwargs):
        self.encoding = encoding
        super(NumpyStringColumn, self).__init__(**kwargs)

    def read_items(self, n_items, buf):
        return np.array(
            buf.read_strings(n_items, encoding=self.encoding), dtype=self.dtype
        )

    def write_items(self, items, buf):
        return buf.write_strings(items.tolist(), encoding=self.encoding)


class NumpyByteStringColumn(NumpyColumn):
    def read_items(self, n_items, buf):
        return np.array(buf.read_strings(n_items), dtype=self.dtype)

    def write_items(self, items, buf):
        return buf.write_strings(items.tolist())


class NumpyFixedString(NumpyStringColumn):
    def __init__(self, length, **kwargs):
        self.length = length
        super(NumpyFixedString, self).__init__(**kwargs)

    def read_items(self, n_items, buf):
        return np.array(buf.read_fixed_strings(
            n_items, self.length, encoding=self.encoding
        ), dtype=self.dtype)

    def write_items(self, items, buf):
        return buf.write_fixed_strings(
            items.tolist(), self.length, encoding=self.encoding
        )


class NumpyByteFixedString(NumpyByteStringColumn):
    def __init__(self, length, **kwargs):
        self.length = length
        super(NumpyByteFixedString, self).__init__(**kwargs)

    def read_items(self, n_items, buf):
        return np.array(
            buf.read_fixed_strings(n_items, self.length), dtype=self.dtype
        )

    def write_items(self, items, buf):
        return buf.write_fixed_strings(items.tolist(), self.length)


def create_string_column(spec, column_options):
    client_settings = column_options['context'].client_settings
    strings_as_bytes = client_settings['strings_as_bytes']
    encoding = client_settings.get(
        'strings_encoding', NumpyStringColumn.default_encoding
    )

    if spec == 'String':
        cls = NumpyByteStringColumn if strings_as_bytes else NumpyStringColumn
        return cls(encoding=encoding, **column_options)
    else:
        length = int(spec[12:-1])
        cls = NumpyByteFixedString if strings_as_bytes else NumpyFixedString
        return cls(length, encoding=encoding, **column_options)



from ... import errors
from ..arraycolumn import create_array_column
from .datecolumn import NumpyDateColumn
from .datetimecolumn import create_numpy_datetime_column
from ..decimalcolumn import create_decimal_column
from ..enumcolumn import create_enum_column
from .floatcolumn import NumpyFloat32Column, NumpyFloat64Column
from .intcolumn import (
    NumpyInt8Column, NumpyInt16Column, NumpyInt32Column, NumpyInt64Column,
    NumpyUInt8Column, NumpyUInt16Column, NumpyUInt32Column, NumpyUInt64Column
)
from .lowcardinalitycolumn import create_numpy_low_cardinality_column
from ..nothingcolumn import NothingColumn
from ..nullcolumn import NullColumn
# from .nullablecolumn import create_nullable_column
from ..simpleaggregatefunctioncolumn import (
    create_simple_aggregate_function_column
)
from .stringcolumn import create_string_column
from ..tuplecolumn import create_tuple_column
from ..uuidcolumn import UUIDColumn
from ..intervalcolumn import (
    IntervalYearColumn, IntervalMonthColumn, IntervalWeekColumn,
    IntervalDayColumn, IntervalHourColumn, IntervalMinuteColumn,
    IntervalSecondColumn
)
from ..ipcolumn import IPv4Column, IPv6Column

column_by_type = {c.ch_type: c for c in [
    NumpyDateColumn,
    NumpyFloat32Column, NumpyFloat64Column,
    NumpyInt8Column, NumpyInt16Column, NumpyInt32Column, NumpyInt64Column,
    NumpyUInt8Column, NumpyUInt16Column, NumpyUInt32Column, NumpyUInt64Column,
    NothingColumn, NullColumn, UUIDColumn,
    IntervalYearColumn, IntervalMonthColumn, IntervalWeekColumn,
    IntervalDayColumn, IntervalHourColumn, IntervalMinuteColumn,
    IntervalSecondColumn, IPv4Column, IPv6Column
]}


def get_numpy_column_by_spec(spec, column_options):
    def create_column_with_options(x):
        return get_numpy_column_by_spec(x, column_options)

    if spec == 'String' or spec.startswith('FixedString'):
        return create_string_column(spec, column_options)

    elif spec.startswith('Enum'):
        return create_enum_column(spec, column_options)

    elif spec.startswith('DateTime'):
        return create_numpy_datetime_column(spec, column_options)

    elif spec.startswith('Decimal'):
        return create_decimal_column(spec, column_options)

    elif spec.startswith('Array'):
        return create_array_column(spec, create_column_with_options)

    elif spec.startswith('Tuple'):
        return create_tuple_column(spec, create_column_with_options)

    # elif spec.startswith('Nullable'):
    #     return create_nullable_column(spec, create_column_with_options)

    elif spec.startswith('LowCardinality'):
        return create_numpy_low_cardinality_column(spec,
                                                   create_column_with_options)

    elif spec.startswith('SimpleAggregateFunction'):
        return create_simple_aggregate_function_column(
            spec, create_column_with_options)

    else:
        try:
            cls = column_by_type[spec]
            return cls(**column_options)

        except KeyError as e:
            raise errors.UnknownTypeError('Unknown type {}'.format(e.args[0]))



import numpy as np

from .base import NumpyColumn


class NumpyDateColumn(NumpyColumn):
    dtype = np.dtype(np.uint16)
    ch_type = 'Date'

    def read_items(self, n_items, buf):
        data = super(NumpyDateColumn, self).read_items(n_items, buf)
        return data.astype('datetime64[D]')



import numpy as np
import pandas as pd


def column_chunks(columns, n):
    for column in columns:
        if not isinstance(column, (np.ndarray, pd.DatetimeIndex)):
            raise TypeError(
                'Unsupported column type: {}. '
                'ndarray/DatetimeIndex is expected.'
                .format(type(column))
            )

    # create chunk generator for every column
    chunked = [
        iter(np.array_split(c, range(0, len(c), n)) if len(c) > n else [c])
        for c in columns
    ]

    while True:
        # get next chunk for every column
        item = [next(column, []) for column in chunked]
        if not any(len(x) for x in item):
            break
        yield item



from collections import namedtuple
from itertools import islice

from ..errors import Error as DriverError
from .errors import InterfaceError, OperationalError, ProgrammingError


Column = namedtuple(
    'Column',
    'name type_code display_size internal_size precision scale null_ok'
)


class Cursor(object):

    class States(object):
        (
            NONE,
            RUNNING,
            FINISHED,
            CURSOR_CLOSED
        ) = range(4)

    _states = States()

    def __init__(self, client, connection):
        self._client = client
        self._connection = connection
        self._reset_state()

        self.arraysize = 1

        # Begin non-PEP attributes
        self._columns_with_types = None
        # End non-PEP attributes

        super(Cursor, self).__init__()

    def __repr__(self):
        is_closed = self._state == self._states.CURSOR_CLOSED
        return '<cursor object at 0x{0:x}; closed: {1:}>'.format(
            id(self), is_closed
        )

    # Iteration support.
    def __iter__(self):
        while True:
            one = self.fetchone()
            if one is None:
                return
            yield one

    # Context manager integrations.
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def description(self):
        if self._state == self._states.NONE:
            return None

        columns = self._columns or []
        types = self._types or []

        return [
            Column(name, type_code, None, None, None, None, True)
            for name, type_code in zip(columns, types)
        ]

    @property
    def rowcount(self):
        """
        :return: the number of rows that the last .execute*() produced.
        """
        return self._rowcount

    def close(self):
        """
        Close the cursor now. The cursor will be unusable from this point
        forward; an :data:`~clickhouse_driver.dbapi.Error` (or subclass)
        exception will be raised if any operation is attempted with the
        cursor.
        """
        self._client.disconnect()
        self._state = self._states.CURSOR_CLOSED

        try:
            # cursor can be already closed
            self._connection.cursors.remove(self)
        except ValueError:
            pass

    def execute(self, operation, parameters=None):
        """
        Prepare and execute a database operation (query or command).

        :param operation: query or command to execute.
        :param parameters: sequence or mapping that will be bound to
                           variables in the operation.
        :return: None
        """
        self._check_cursor_closed()
        self._begin_query()

        try:
            execute, execute_kwargs = self._prepare()

            response = execute(
                operation, params=parameters, with_column_types=True,
                **execute_kwargs
            )

        except DriverError as orig:
            raise OperationalError(orig)

        self._process_response(response)
        self._end_query()

    def executemany(self, operation, seq_of_parameters):
        """
        Prepare a database operation (query or command) and then execute it
        against all parameter sequences found in the sequence
        `seq_of_parameters`.

        :param operation: query or command to execute.
        :param seq_of_parameters: sequences or mappings for execution.
        :return: None
        """
        self._check_cursor_closed()
        self._begin_query()

        try:
            execute, execute_kwargs = self._prepare()

            response = execute(
                operation, params=seq_of_parameters, **execute_kwargs
            )

        except DriverError as orig:
            raise OperationalError(orig)

        self._process_response(response, executemany=True)
        self._end_query()

    def fetchone(self):
        """
        Fetch the next row of a query result set, returning a single sequence,
        or None when no more data is available.

        :return: the next row of a query result set or None.
        """
        self._check_query_started()

        if self._stream_results:
            return next(self._rows, None)

        else:
            if not self._rows:
                return None

            return self._rows.pop(0)

    def fetchmany(self, size=None):
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.

        :param size: amount of rows to return.
        :return: list of fetched rows or empty list.
        """
        self._check_query_started()

        if size is None:
            size = self.arraysize

        if self._stream_results:
            if size == -1:
                return list(self._rows)
            else:
                return list(islice(self._rows, size))

        if size < 0:
            rv = self._rows
            self._rows = []
        else:
            rv = self._rows[:size]
            self._rows = self._rows[size:]

        return rv

    def fetchall(self):
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples).

        :return: list of fetched rows.
        """
        self._check_query_started()

        if self._stream_results:
            return list(self._rows)

        rv = self._rows
        self._rows = []
        return rv

    def setinputsizes(self, sizes):
        # Do nothing.
        pass

    def setoutputsize(self, size, column=None):
        # Do nothing.
        pass

    # Begin non-PEP methods
    @property
    def columns_with_types(self):
        """
        :return: list of column names with corresponding types of the last
                 .execute*(). E.g. [('x', 'UInt64')].
        """
        return self._columns_with_types

    def set_stream_results(self, stream_results, max_row_buffer):
        """
        Toggles results streaming from server. Driver will consume
        block-by-block of `max_row_buffer` size and yield row-by-row from each
        block.

        :param stream_results: enable or disable results streaming.
        :param max_row_buffer: specifies the maximum number of rows to buffer
               at a time.
        :return: None
        """
        self._stream_results = stream_results
        self._max_row_buffer = max_row_buffer

    def set_settings(self, settings):
        """
        Specifies settings for cursor.

        :param settings: dictionary of query settings
        :return: None
        """
        self._settings = settings

    def set_types_check(self, types_check):
        """
        Toggles type checking for sequence of INSERT parameters.
        Disabled by default.

        :param types_check: new types check value.
        :return: None
        """
        self._types_check = types_check

    def set_external_table(self, name, structure, data):
        """
        Adds external table to cursor context.

        If the same table is specified more than once the last one is used.

        :param name: name of external table
        :param structure: list of tuples (name, type) that defines table
                          structure. Example [(x, 'Int32')].
        :param data: sequence of rows of tuples or dicts for transmission.
        :return: None
        """
        self._external_tables[name] = (structure, data)

    def set_query_id(self, query_id):
        """
        Specifies the query identifier for cursor.

        :param query_id: the query identifier.
        :return: None
        """
        self._query_id = query_id
    # End non-PEP methods

    # Private methods.
    def _prepare(self):
        external_tables = [
            {'name': name, 'structure': structure, 'data': data}
            for name, (structure, data) in self._external_tables.items()
        ] or None

        execute = self._client.execute

        if self._stream_results:
            execute = self._client.execute_iter
            self._settings = self._settings or {}
            self._settings['max_block_size'] = self._max_row_buffer

        execute_kwargs = {
            'settings': self._settings,
            'external_tables': external_tables,
            'types_check': self._types_check,
            'query_id': self._query_id
        }

        return execute, execute_kwargs

    def _process_response(self, response, executemany=False):
        if executemany:
            self._rowcount = response
            response = None

        if not response:
            self._columns = self._types = self._rows = []
            return

        if self._stream_results:
            columns_with_types = next(response)
            rows = response

        else:
            rows, columns_with_types = response

        self._columns_with_types = columns_with_types

        # Only SELECT queries have columns_with_types.
        # DDL and INSERT INTO ... SELECT queries have empty columns header.
        # We need to obtain rows count only during non-streaming SELECTs.
        if columns_with_types:
            self._columns, self._types = zip(*columns_with_types)
            if not self._stream_results:
                self._rowcount = len(rows)
        else:
            self._columns = self._types = []

        self._rows = rows

    def _reset_state(self):
        """
        Resets query state and get ready for another query.
        """
        self._state = self._states.NONE

        self._columns = None
        self._types = None
        self._rows = None
        self._rowcount = -1

        self._stream_results = False
        self._max_row_buffer = 0
        self._settings = None
        self._query_id = None
        self._external_tables = {}
        self._types_check = False

    def _begin_query(self):
        self._state = self._states.RUNNING

    def _end_query(self):
        self._state = self._states.FINISHED

    def _check_cursor_closed(self):
        if self._state == self._states.CURSOR_CLOSED:
            raise InterfaceError('cursor already closed')

    def _check_query_started(self):
        if self._state == self._states.NONE:
            raise ProgrammingError('no results to fetch')



from .types import (
    SettingUInt64, SettingBool, SettingFloat, SettingString, SettingMaxThreads,
    SettingChar
)

SettingInt64 = SettingUInt64

# Seconds and milliseconds should be set as ints.
SettingSeconds = SettingMilliseconds = SettingUInt64

# Server cares about possible choices validation.
# See https://github.com/yandex/ClickHouse/blob/master/dbms/src/
# Interpreters/Settings.h for all choices.
SettingLoadBalancing = SettingTotalsMode = SettingCompressionMethod = \
    SettingDistributedProductMode = SettingGlobalSubqueriesMethod = \
    SettingDateTimeInputFormat = \
    SettingURI = \
    SettingJoinAlgorithm = \
    SettingSpecialSort = \
    SettingLogQueriesType = \
    SettingDefaultDatabaseEngine = \
    SettingString

settings = {
    # Settings
    'min_compress_block_size': SettingUInt64,
    'max_compress_block_size': SettingUInt64,
    'max_block_size': SettingUInt64,
    'max_insert_block_size': SettingUInt64,
    'min_insert_block_size_rows': SettingUInt64,
    'min_insert_block_size_bytes': SettingUInt64,
    'max_partitions_per_insert_block': SettingUInt64,
    'max_threads': SettingMaxThreads,
    'max_alter_threads': SettingMaxThreads,
    'max_read_buffer_size': SettingUInt64,
    'max_distributed_connections': SettingUInt64,
    'max_query_size': SettingUInt64,
    'interactive_delay': SettingUInt64,
    'connect_timeout': SettingSeconds,
    'connect_timeout_with_failover_ms': SettingMilliseconds,
    'receive_timeout': SettingSeconds,
    'send_timeout': SettingSeconds,
    'queue_max_wait_ms': SettingMilliseconds,
    'poll_interval': SettingUInt64,
    'distributed_connections_pool_size': SettingUInt64,
    'connections_with_failover_max_tries': SettingUInt64,
    'extremes': SettingBool,
    'use_uncompressed_cache': SettingBool,
    'replace_running_query': SettingBool,
    'background_pool_size': SettingUInt64,
    'background_schedule_pool_size': SettingUInt64,

    'distributed_directory_monitor_sleep_time_ms': SettingMilliseconds,

    'distributed_directory_monitor_batch_inserts': SettingBool,
    'distributed_directory_monitor_max_sleep_time_ms': SettingMilliseconds,

    'optimize_move_to_prewhere': SettingBool,
    'optimize_skip_unused_shards': SettingBool,
    'optimize_read_in_order': SettingBool,
    'optimize_min_equality_disjunction_chain_length': SettingUInt64,
    'enable_optimize_predicate_expression': SettingBool,

    'replication_alter_partitions_sync': SettingUInt64,
    'replication_alter_columns_timeout': SettingUInt64,

    'load_balancing': SettingLoadBalancing,

    'totals_mode': SettingTotalsMode,
    'totals_auto_threshold': SettingFloat,

    'compile': SettingBool,
    'compile_expressions': SettingBool,
    'min_count_to_compile': SettingUInt64,
    'group_by_two_level_threshold': SettingUInt64,
    'group_by_two_level_threshold_bytes': SettingUInt64,
    'distributed_aggregation_memory_efficient': SettingBool,
    'aggregation_memory_efficient_merge_threads': SettingUInt64,

    'max_parallel_replicas': SettingUInt64,
    'parallel_replicas_count': SettingUInt64,
    'parallel_replica_offset': SettingUInt64,

    'skip_unavailable_shards': SettingBool,

    'distributed_group_by_no_merge': SettingBool,

    'merge_tree_min_rows_for_concurrent_read': SettingUInt64,
    'merge_tree_min_rows_for_seek': SettingUInt64,
    'merge_tree_coarse_index_granularity': SettingUInt64,
    'merge_tree_max_rows_to_use_cache': SettingUInt64,
    'merge_tree_min_bytes_for_concurrent_read': SettingUInt64,
    'merge_tree_min_bytes_for_seek': SettingUInt64,
    'merge_tree_max_bytes_to_use_cache': SettingUInt64,
    'merge_tree_uniform_read_distribution': SettingBool,

    'mysql_max_rows_to_insert': SettingUInt64,

    'min_bytes_to_use_direct_io': SettingUInt64,

    'force_index_by_date': SettingBool,
    'force_primary_key': SettingBool,

    'mark_cache_min_lifetime': SettingUInt64,

    'max_streams_to_max_threads_ratio': SettingFloat,

    'network_compression_method': SettingCompressionMethod,

    'network_zstd_compression_level': SettingInt64,

    'priority': SettingUInt64,

    'log_queries': SettingBool,
    'log_queries_cut_to_length': SettingUInt64,
    'query_profiler_real_time_period_ns': SettingUInt64,
    'query_profiler_cpu_time_period_ns': SettingUInt64,
    'enable_debug_queries': SettingBool,

    'distributed_product_mode': SettingDistributedProductMode,

    'max_concurrent_queries_for_user': SettingUInt64,

    'insert_deduplicate': SettingBool,

    'insert_quorum': SettingUInt64,
    'insert_quorum_timeout': SettingMilliseconds,
    'select_sequential_consistency': SettingUInt64,
    'table_function_remote_max_addresses': SettingUInt64,
    'read_backoff_min_latency_ms': SettingMilliseconds,
    'read_backoff_max_throughput': SettingUInt64,
    'read_backoff_min_interval_between_events_ms': SettingMilliseconds,
    'read_backoff_min_events': SettingUInt64,

    'memory_tracker_fault_probability': SettingFloat,

    'enable_http_compression': SettingBool,
    'http_zlib_compression_level': SettingInt64,

    'http_native_compression_disable_checksumming_on_decompress': SettingBool,

    'count_distinct_implementation': SettingString,

    'add_http_cors_header': SettingBool,

    'input_format_skip_unknown_fields': SettingBool,
    'input_format_import_nested_json': SettingBool,
    'input_format_values_interpret_expressions': SettingBool,
    'input_format_with_names_use_header': SettingBool,
    'input_format_defaults_for_omitted_fields': SettingBool,
    'input_format_null_as_default': SettingBool,
    'input_format_values_deduce_templates_of_expressions': SettingBool,
    'input_format_values_accurate_types_of_literals': SettingBool,
    'input_format_allow_errors_num': SettingUInt64,
    'input_format_allow_errors_ratio': SettingFloat,
    'input_format_csv_unquoted_null_literal_as_null': SettingBool,

    'output_format_write_statistics': SettingBool,
    'output_format_json_quote_64bit_integers': SettingBool,
    'output_format_json_quote_denormals': SettingBool,
    'output_format_json_escape_forward_slashes': SettingBool,
    'output_format_pretty_max_rows': SettingUInt64,
    'output_format_pretty_max_column_pad_width': SettingUInt64,
    'output_format_pretty_color': SettingBool,
    'output_format_parquet_row_group_size': SettingUInt64,

    'use_client_time_zone': SettingBool,

    'send_progress_in_http_headers': SettingBool,

    'http_headers_progress_interval_ms': SettingUInt64,

    'fsync_metadata': SettingBool,

    'join_use_nulls': SettingBool,
    'join_default_strictness': SettingString,
    'preferred_block_size_bytes': SettingUInt64,

    'max_replica_delay_for_distributed_queries': SettingUInt64,
    'fallback_to_stale_replicas_for_distributed_queries': SettingBool,
    'preferred_max_column_in_block_size_bytes': SettingUInt64,

    'insert_distributed_sync': SettingBool,
    'insert_distributed_timeout': SettingUInt64,
    'distributed_ddl_task_timeout': SettingInt64,
    'stream_flush_interval_ms': SettingMilliseconds,
    'format_schema': SettingString,
    'insert_allow_materialized_columns': SettingBool,
    'http_connection_timeout': SettingSeconds,
    'http_send_timeout': SettingSeconds,
    'http_receive_timeout': SettingSeconds,
    'optimize_throw_if_noop': SettingBool,
    'use_index_for_in_with_subqueries': SettingBool,
    'empty_result_for_aggregation_by_empty_set': SettingBool,
    'allow_distributed_ddl': SettingBool,
    'odbc_max_field_size': SettingUInt64,

    # Limits
    'max_rows_to_read': SettingUInt64,
    'max_bytes_to_read': SettingUInt64,
    'read_overflow_mode': SettingString,

    'max_rows_to_group_by': SettingUInt64,
    'group_by_overflow_mode': SettingString,
    'max_bytes_before_external_group_by': SettingUInt64,

    'max_rows_to_sort': SettingUInt64,
    'max_bytes_to_sort': SettingUInt64,
    'sort_overflow_mode': SettingString,
    'max_bytes_before_external_sort': SettingUInt64,
    'max_bytes_before_remerge_sort': SettingUInt64,
    'max_result_rows': SettingUInt64,
    'max_result_bytes': SettingUInt64,
    'result_overflow_mode': SettingString,

    'max_execution_time': SettingSeconds,
    'timeout_overflow_mode': SettingString,

    'min_execution_speed': SettingUInt64,
    'timeout_before_checking_execution_speed': SettingSeconds,

    'max_columns_to_read': SettingUInt64,
    'max_temporary_columns': SettingUInt64,
    'max_temporary_non_const_columns': SettingUInt64,

    'max_subquery_depth': SettingUInt64,
    'max_pipeline_depth': SettingUInt64,
    'max_ast_depth': SettingUInt64,
    'max_ast_elements': SettingUInt64,
    'max_expanded_ast_elements': SettingUInt64,

    'readonly': SettingUInt64,

    'max_rows_in_set': SettingUInt64,
    'max_bytes_in_set': SettingUInt64,
    'set_overflow_mode': SettingString,

    'max_rows_in_join': SettingUInt64,
    'max_bytes_in_join': SettingUInt64,
    'join_overflow_mode': SettingString,

    'max_rows_to_transfer': SettingUInt64,
    'max_bytes_to_transfer': SettingUInt64,
    'transfer_overflow_mode': SettingString,

    'max_rows_in_distinct': SettingUInt64,
    'max_bytes_in_distinct': SettingUInt64,
    'distinct_overflow_mode': SettingString,

    'max_memory_usage': SettingUInt64,
    'max_memory_usage_for_user': SettingUInt64,
    'max_memory_usage_for_all_queries': SettingUInt64,

    'max_network_bandwidth': SettingUInt64,
    'max_network_bytes': SettingUInt64,
    'max_network_bandwidth_for_user': SettingUInt64,
    'max_network_bandwidth_for_all_users': SettingUInt64,

    'max_streams_multiplier_for_merge_tables': SettingFloat,
    'max_http_get_redirects': SettingUInt64,
    'max_execution_speed': SettingUInt64,
    'max_execution_speed_bytes': SettingUInt64,

    'format_csv_delimiter': SettingChar,
    'format_csv_allow_single_quotes': SettingBool,
    'format_csv_allow_double_quotes': SettingBool,

    'format_template_resultset': SettingString,
    'format_template_row': SettingString,
    'format_template_rows_between_delimiter': SettingString,
    'format_custom_escaping_rule': SettingString,
    'format_custom_field_delimiter': SettingString,
    'format_custom_row_before_delimiter': SettingString,
    'format_custom_row_after_delimiter': SettingString,
    'format_custom_row_between_delimiter': SettingString,
    'format_custom_result_before_delimiter': SettingString,
    'format_custom_result_after_delimiter': SettingString,

    'enable_conditional_computation': SettingUInt64,

    'date_time_input_format': SettingDateTimeInputFormat,
    'log_profile_events': SettingBool,
    'log_query_settings': SettingBool,
    'log_query_threads': SettingBool,
    'send_logs_level': SettingString,
    'low_cardinality_max_dictionary_size': SettingUInt64,
    'low_cardinality_use_single_dictionary_for_part': SettingBool,
    'decimal_check_overflow': SettingBool,
    'prefer_localhost_replica': SettingBool,
    'max_fetch_partition_retries_count': SettingUInt64,
    'asterisk_left_columns_only': SettingBool,
    'http_max_multipart_form_data_size': SettingUInt64,
    'calculate_text_stack_trace': SettingBool,
    'parallel_view_processing': SettingBool,

    'allow_experimental_low_cardinality_type': SettingBool,
    'allow_experimental_decimal_type': SettingBool,
    'allow_suspicious_low_cardinality_types': SettingBool,
    'allow_experimental_multiple_joins_emulation': SettingBool,
    'allow_experimental_cross_to_join_conversion': SettingBool,
    'allow_experimental_data_skipping_indices': SettingBool,
    'allow_hyperscan': SettingBool,
    'allow_simdjson': SettingBool,
    'allow_introspection_functions': SettingBool,
    'allow_drop_detached': SettingBool,
    'allow_experimental_live_view': SettingBool,
    'allow_ddl': SettingBool,

    'partial_merge_join': SettingBool,
    'partial_merge_join_optimizations': SettingBool,
    'partial_merge_join_rows_in_right_blocks': SettingUInt64,
    'partial_merge_join_rows_in_left_blocks': SettingFloat,

    'distributed_replica_error_half_life': SettingSeconds,
    'distributed_replica_error_cap': SettingUInt64,

    'min_free_disk_space_for_temporary_data': SettingUInt64,
    'tcp_keep_alive_timeout': SettingSeconds,
    'connection_pool_max_wait_ms': SettingMilliseconds,
    'kafka_max_wait_ms': SettingMilliseconds,
    'idle_connection_timeout': SettingUInt64,
    's3_min_upload_part_size': SettingUInt64,
    'any_join_distinct_right_table_keys': SettingBool,
    'join_any_take_last_row': SettingBool,
    'stream_poll_timeout_ms': SettingMilliseconds,
    'joined_subquery_requires_alias': SettingBool,
    'enable_unaligned_array_join': SettingBool,
    'low_cardinality_allow_in_native_format': SettingBool,
    'external_table_functions_use_nulls': SettingBool,
    'experimental_use_processors': SettingBool,
    'check_query_single_value_result': SettingBool,
    'live_view_heartbeat_interval': SettingSeconds,
    'temporary_live_view_timeout': SettingSeconds,
    'max_live_view_insert_blocks_before_refresh': SettingUInt64,

    'max_insert_threads': SettingUInt64,
    'replace_running_query_max_wait_ms': SettingMilliseconds,
    'background_move_pool_size': SettingUInt64,
    'min_count_to_compile_expression': SettingUInt64,
    'force_optimize_skip_unused_shards': SettingUInt64,
    'input_format_parallel_parsing': SettingBool,
    'min_chunk_bytes_for_parallel_parsing': SettingUInt64,
    'min_bytes_to_use_mmap_io': SettingUInt64,
    'os_thread_priority': SettingInt64,
    'input_format_tsv_empty_as_default': SettingBool,
    'input_format_avro_schema_registry_url': SettingString,
    'output_format_avro_codec': SettingString,
    'output_format_avro_sync_interval': SettingUInt64,
    'min_execution_speed_bytes': SettingUInt64,
    'default_max_bytes_in_join': SettingUInt64,
    'enable_optimize_predicate_expression_to_final_subquery': SettingBool,
    'cancel_http_readonly_queries_on_client_close': SettingBool,
    'enable_scalar_subquery_optimization': SettingBool,
    'optimize_trivial_count_query': SettingBool,
    'mutations_sync': SettingUInt64,
    'optimize_if_chain_to_miltiif': SettingBool,
    'max_parser_depth': SettingUInt64,

    'max_joined_block_size_rows': SettingUInt64,
    'connect_timeout_with_failover_secure_ms': SettingMilliseconds,
    'parallel_distributed_insert_select': SettingBool,
    'force_optimize_skip_unused_shards_no_nested': SettingBool,
    'format_avro_schema_registry_url': SettingURI,
    'output_format_tsv_crlf_end_of_line': SettingBool,
    'join_algorithm': SettingJoinAlgorithm,
    'memory_profiler_step': SettingUInt64,
    'output_format_csv_crlf_end_of_line': SettingBool,
    'allow_experimental_alter_materialized_view_structure': SettingBool,
    'enable_early_constant_folding': SettingBool,
    'deduplicate_blocks_in_dependent_materialized_views': SettingBool,
    'use_compact_format_in_distributed_parts_names': SettingBool,
    'multiple_joins_rewriter_version': SettingUInt64,

    'min_insert_block_size_rows_for_materialized_views': SettingUInt64,
    'min_insert_block_size_bytes_for_materialized_views': SettingUInt64,
    'max_final_threads': SettingUInt64,
    'background_buffer_flush_schedule_pool_size': SettingUInt64,
    'background_distributed_schedule_pool_size': SettingUInt64,
    'special_sort': SettingSpecialSort,
    'optimize_distributed_group_by_sharding_key': SettingBool,
    'log_queries_min_type': SettingLogQueriesType,
    'allow_suspicious_codecs': SettingBool,
    'metrics_perf_events_enabled': SettingBool,
    'metrics_perf_events_list': SettingString,
    'join_on_disk_max_files_to_merge': SettingUInt64,
    'temporary_files_codec': SettingString,
    'max_untracked_memory': SettingUInt64,
    'memory_profiler_sample_probability': SettingFloat,
    'optimize_aggregation_in_order': SettingBool,
    'default_database_engine': SettingDefaultDatabaseEngine,
    'allow_experimental_database_atomic': SettingBool,
    'show_table_uuid_in_table_create_query_if_not_nil': SettingBool,
    'optimize_arithmetic_operations_in_aggregate_functions': SettingBool,
    'validate_polygons': SettingBool,
    'transform_null_in': SettingBool,
    'allow_nondeterministic_mutations': SettingBool,
    'lock_acquire_timeout': SettingSeconds,
    'materialize_ttl_after_modify': SettingBool,
    'allow_experimental_geo_types': SettingBool,
    'output_format_pretty_max_value_width': SettingUInt64,
    'format_regexp': SettingString,
    'format_regexp_escaping_rule': SettingString,
    'format_regexp_skip_unmatched': SettingBool,
    'output_format_enable_streaming': SettingBool,
}



from decimal import Decimal, localcontext

from .base import FormatColumn
from .exceptions import ColumnTypeMismatchException
from .intcolumn import Int128Column, Int256Column


class DecimalColumn(FormatColumn):
    py_types = (Decimal, float, int)
    max_precision = None
    int_size = None

    def __init__(self, precision, scale, types_check=False, **kwargs):
        self.precision = precision
        self.scale = scale
        super(DecimalColumn, self).__init__(**kwargs)

        if types_check:
            max_signed_int = (1 << (8 * self.int_size - 1)) - 1

            def check_item(value):
                if value < -max_signed_int or value > max_signed_int:
                    raise ColumnTypeMismatchException(value)

            self.check_item = check_item

    def after_read_items(self, items, nulls_map=None):
        if self.scale >= 1:
            scale = 10 ** self.scale

            if nulls_map is None:
                return tuple(Decimal(item) / scale for item in items)
            else:
                return tuple(
                    (None if is_null else Decimal(items[i]) / scale)
                    for i, is_null in enumerate(nulls_map)
                )
        else:
            if nulls_map is None:
                return tuple(Decimal(item) for item in items)
            else:
                return tuple(
                    (None if is_null else Decimal(items[i]))
                    for i, is_null in enumerate(nulls_map)
                )

    def before_write_items(self, items, nulls_map=None):
        null_value = self.null_value

        if self.scale >= 1:
            scale = 10 ** self.scale

            for i, item in enumerate(items):
                if nulls_map and nulls_map[i]:
                    items[i] = null_value
                else:
                    items[i] = int(Decimal(str(item)) * scale)

        else:
            for i, item in enumerate(items):
                if nulls_map and nulls_map[i]:
                    items[i] = null_value
                else:
                    items[i] = int(Decimal(str(item)))

    # Override default precision to the maximum supported by underlying type.
    def _write_data(self, items, buf):
        with localcontext() as ctx:
            ctx.prec = self.max_precision
            super(DecimalColumn, self)._write_data(items, buf)

    def _read_data(self, n_items, buf, nulls_map=None):
        with localcontext() as ctx:
            ctx.prec = self.max_precision
            return super(DecimalColumn, self)._read_data(
                n_items, buf, nulls_map=nulls_map
            )


class Decimal32Column(DecimalColumn):
    format = 'i'
    max_precision = 9
    int_size = 4


class Decimal64Column(DecimalColumn):
    format = 'q'
    max_precision = 18
    int_size = 8


class Decimal128Column(DecimalColumn, Int128Column):
    max_precision = 38


class Decimal256Column(DecimalColumn, Int256Column):
    max_precision = 76


def create_decimal_column(spec, column_options):
    precision, scale = spec[8:-1].split(',')
    precision, scale = int(precision), int(scale)

    # Maximum precisions for underlying types are:
    # Int32    10**9
    # Int64   10**18
    # Int128  10**38
    # Int256  10**76
    if precision <= 9:
        cls = Decimal32Column
    elif precision <= 18:
        cls = Decimal64Column
    elif precision <= 38:
        cls = Decimal128Column
    else:
        cls = Decimal256Column

    return cls(precision, scale, **column_options)





def create_simple_aggregate_function_column(spec, column_by_spec_getter):
    # SimpleAggregateFunction(Func, Type) -> Type
    inner = spec[24:-1].split(',', 1)[1].strip()
    nested = column_by_spec_getter(inner)
    return nested




from .. import defines
from .base import Column


class String(Column):
    ch_type = 'String'
    py_types = (str, )
    null_value = ''

    default_encoding = defines.STRINGS_ENCODING

    def __init__(self, encoding=default_encoding, **kwargs):
        self.encoding = encoding
        super(String, self).__init__(**kwargs)

    def write_items(self, items, buf):
        buf.write_strings(items, encoding=self.encoding)

    def read_items(self, n_items, buf):
        return buf.read_strings(n_items, encoding=self.encoding)


class ByteString(String):
    py_types = (bytes, )
    null_value = b''

    def write_items(self, items, buf):
        buf.write_strings(items)

    def read_items(self, n_items, buf):
        return buf.read_strings(n_items)


class FixedString(String):
    ch_type = 'FixedString'

    def __init__(self, length, **kwargs):
        self.length = length
        super(FixedString, self).__init__(**kwargs)

    def read_items(self, n_items, buf):
        return buf.read_fixed_strings(
            n_items, self.length, encoding=self.encoding
        )

    def write_items(self, items, buf):
        buf.write_fixed_strings(items, self.length, encoding=self.encoding)


class ByteFixedString(FixedString):
    py_types = (bytearray, bytes)
    null_value = b''

    def read_items(self, n_items, buf):
        return buf.read_fixed_strings(n_items, self.length)

    def write_items(self, items, buf):
        buf.write_fixed_strings(items, self.length)


def create_string_column(spec, column_options):
    client_settings = column_options['context'].client_settings
    strings_as_bytes = client_settings['strings_as_bytes']
    encoding = client_settings.get('strings_encoding', String.default_encoding)

    if spec == 'String':
        cls = ByteString if strings_as_bytes else String
        return cls(encoding=encoding, **column_options)
    else:
        length = int(spec[12:-1])
        cls = ByteFixedString if strings_as_bytes else FixedString
        return cls(length, encoding=encoding, **column_options)



from datetime import datetime

from pytz import timezone as get_timezone, utc
from tzlocal import get_localzone

from .base import FormatColumn

EPOCH = datetime(1970, 1, 1, tzinfo=utc)


class DateTimeColumn(FormatColumn):
    ch_type = 'DateTime'
    py_types = (datetime, int)
    format = 'I'

    def __init__(self, timezone=None, offset_naive=True, **kwargs):
        self.timezone = timezone
        self.offset_naive = offset_naive
        super(DateTimeColumn, self).__init__(**kwargs)

    def after_read_items(self, items, nulls_map=None):
        tz = self.timezone
        fromts = datetime.fromtimestamp

        # A bit ugly copy-paste. But it helps save time on items
        # processing by avoiding lambda calls or if in loop.
        if self.offset_naive:
            if tz:
                if nulls_map is None:
                    return tuple(
                        fromts(item, tz).replace(tzinfo=None)
                        for item in items
                    )
                else:
                    return tuple(
                        (None if is_null else
                         fromts(items[i], tz).replace(tzinfo=None))
                        for i, is_null in enumerate(nulls_map)
                    )
            else:
                if nulls_map is None:
                    return tuple(fromts(item) for item in items)
                else:
                    return tuple(
                        (None if is_null else fromts(items[i]))
                        for i, is_null in enumerate(nulls_map)
                    )

        else:
            if nulls_map is None:
                return tuple(fromts(item, tz) for item in items)
            else:
                return tuple(
                    (None if is_null else fromts(items[i], tz))
                    for i, is_null in enumerate(nulls_map)
                )

    def before_write_items(self, items, nulls_map=None):
        timezone = self.timezone
        null_value = self.null_value
        to_timestamp = datetime.timestamp

        for i, item in enumerate(items):
            if nulls_map and nulls_map[i]:
                items[i] = null_value
                continue

            if isinstance(item, int):
                # support supplying raw integers to avoid
                # costly timezone conversions when using datetime
                continue

            if timezone:
                # Set server's timezone for offset-naive datetime.
                if item.tzinfo is None:
                    item = timezone.localize(item)

                item = item.astimezone(utc)

            else:
                # If datetime is offset-aware use it's timezone.
                if item.tzinfo is not None:
                    item = item.astimezone(utc)

            items[i] = int(to_timestamp(item))


class DateTime64Column(DateTimeColumn):
    ch_type = 'DateTime64'
    format = 'Q'

    max_scale = 6

    def __init__(self, scale=0, **kwargs):
        self.scale = scale
        super(DateTime64Column, self).__init__(**kwargs)

    def after_read_items(self, items, nulls_map=None):
        scale = float(10 ** self.scale)

        tz = self.timezone
        fromts = datetime.fromtimestamp

        # A bit ugly copy-paste. But it helps save time on items
        # processing by avoiding lambda calls or if in loop.
        if self.offset_naive:
            if tz:
                if nulls_map is None:
                    return tuple(
                        fromts(item / scale, tz).replace(tzinfo=None)
                        for item in items
                    )
                else:
                    return tuple(
                        (None if is_null else
                         fromts(items[i] / scale, tz).replace(tzinfo=None))
                        for i, is_null in enumerate(nulls_map)
                    )
            else:
                if nulls_map is None:
                    return tuple(fromts(item / scale) for item in items)
                else:
                    return tuple(
                        (None if is_null else fromts(items[i] / scale))
                        for i, is_null in enumerate(nulls_map)
                    )

        else:
            if nulls_map is None:
                return tuple(fromts(item / scale, tz) for item in items)
            else:
                return tuple(
                    (None if is_null else fromts(items[i] / scale, tz))
                    for i, is_null in enumerate(nulls_map)
                )

    def before_write_items(self, items, nulls_map=None):
        scale = 10 ** self.scale
        frac_scale = 10 ** (self.max_scale - self.scale)

        timezone = self.timezone
        null_value = self.null_value
        to_timestamp = datetime.timestamp

        for i, item in enumerate(items):
            if nulls_map and nulls_map[i]:
                items[i] = null_value
                continue

            if isinstance(item, int):
                # support supplying raw integers to avoid
                # costly timezone conversions when using datetime
                continue

            if timezone:
                # Set server's timezone for offset-naive datetime.
                if item.tzinfo is None:
                    item = timezone.localize(item)

                item = item.astimezone(utc)

            else:
                # If datetime is offset-aware use it's timezone.
                if item.tzinfo is not None:
                    item = item.astimezone(utc)

            items[i] = (
                int(to_timestamp(item)) * scale +
                int(item.microsecond / frac_scale)
            )


def create_datetime_column(spec, column_options):
    if spec.startswith('DateTime64'):
        cls = DateTime64Column
        spec = spec[11:-1]
        params = spec.split(',', 1)
        column_options['scale'] = int(params[0])
        if len(params) > 1:
            spec = params[1].strip() + ')'
    else:
        cls = DateTimeColumn
        spec = spec[9:]

    context = column_options['context']

    tz_name = timezone = None
    offset_naive = True

    # Use column's timezone if it's specified.
    if spec and spec[-1] == ')':
        tz_name = spec[1:-2]
        offset_naive = False
    else:
        if not context.settings.get('use_client_time_zone', False):
            try:
                local_timezone = get_localzone().zone
            except Exception:
                local_timezone = None

            if local_timezone != context.server_info.timezone:
                tz_name = context.server_info.timezone

    if tz_name:
        timezone = get_timezone(tz_name)

    return cls(timezone=timezone, offset_naive=offset_naive, **column_options)
